import re
import xml.etree.ElementTree as etree

from markdown.blockprocessors import BlockProcessor
from markdown.extensions import Extension


class NydokRequirementProcessor(BlockProcessor):
    """Process nydok requirements in format:

    - FR001 [UR001]: The system must be able to do something
    """

    REQ_PATTERN = r"- (?P<req_id>[A-Z]+[0-9]+)( \[(?P<refs>[A-Z,0-9]+)\])?: (?P<desc>.*)"

    def __init__(self, parser):
        super().__init__(parser)
        self.RE = re.compile(self.REQ_PATTERN)

    def test(self, parent, block):
        return bool(self.RE.match(block))

    def run(self, parent, blocks):
        # Check fr multiple items in one block.
        items = self.get_items(blocks.pop(0))

        # This is a new list so create parent with appropriate tag.
        req_table = etree.SubElement(parent, "table")
        req_table.attrib["class"] = "nydok-requirements"

        # Loop through items in block, recursively parsing each with the
        # appropriate parent.

        has_refs = any(item.get("refs") for item in items)
        for item in items:
            # New item. Create li and parse with it as parent
            tr = etree.SubElement(req_table, "tr")
            req_td = etree.SubElement(tr, "td")
            req_td.text = item["req_id"]
            req_td.attrib["class"] = "req-id"
            if has_refs:
                ref_td = etree.SubElement(tr, "td")
                ref_td.text = " ".join(item["refs"].split(","))
                ref_td.attrib["class"] = "req-refs"
            desc_td = etree.SubElement(tr, "td")
            desc_td.text = item["desc"]
            desc_td.attrib["class"] = "req-desc"

    def get_items(self, block):
        """Break a block into list items."""
        items = []
        for line in block.split("\n"):
            m = self.RE.match(line)
            if m:
                items.append(m.groupdict())
        return items


class NydokExtension(Extension):
    """Nydok extension for Python-Markdown."""

    def extendMarkdown(self, md):
        md.parser.blockprocessors.register(NydokRequirementProcessor(md.parser), "nydok", 100)
        md.registerExtension(self)



import re
from pathlib import Path
from typing import Optional

import pytest

from ..exception import (
    DuplicateRequirementException,
    FailedTestCaseException,
    MissingTestCaseException,
)
from ..schema import Requirement
from .specsmanager import specs_manager

DEFAULT_REQ_PATTERN = r"- (?P<req_id>[A-Z]+[0-9]+)( \[(?P<refs>[A-Z,0-9]+)\])?: (?P<desc>.*)"

RE_NEW_LINE = re.compile(r"\n")


class SpecFile(pytest.File):
    def collect(self):
        with open(self.fspath) as f:
            md_data = f.read()

        # Get regex to use
        req_regex = (
            self.config.getoption("nydok-specs-regex")
            or self.config.getini("nydok-specs-regex")
            or DEFAULT_REQ_PATTERN
        )

        # Parse out all requirements from file, and yield py.test Items
        pattern = re.compile(req_regex)
        for m in pattern.finditer(md_data):
            line_no = len(RE_NEW_LINE.findall(md_data, 0, m.start(1))) + 1
            matches = m.groupdict()
            req_id, desc = matches["req_id"], matches["desc"]
            refs = []
            if matches.get("refs"):
                refs = matches["refs"].split(",")
            req = Requirement(req_id, desc, Path(self.fspath), line_no, refs)
            specs_manager.add_requirement(req)
            yield ReqItem.from_parent(self, name=req_id, req=req)


class ReqItem(pytest.Item):
    def __init__(self, name, parent, req: Requirement):
        super().__init__(name, parent)
        self.req: Requirement = req
        self.duplicate_item: Optional[pytest.Item] = (
            None  # If there are duplicated Items with same name
        )

    def runtest(self):
        if self.duplicate_item:
            raise DuplicateRequirementException()
        if not specs_manager.has_test_case(self.name):
            raise MissingTestCaseException()
        if not specs_manager.requirement_passed(self.name):
            raise FailedTestCaseException()

    def _func_name_suggestion(self):
        return "_".join(self.req.desc.lower().split(" ")[:8])

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        req_info = f"{self.name} ({self.req.file_path.resolve()}:{self.req.line_no})"

        if isinstance(excinfo.value, DuplicateRequirementException):
            req_duplicate_info = f"{self.name} ({self.req.file_path}:{self.req.line_no})"
            return f"Requirement {req_info} is already defined in {req_duplicate_info}\n"

        if isinstance(excinfo.value, MissingTestCaseException):
            receipe = (
                f'@testcase("{self.name}")\n'
                f"def test_{self._func_name_suggestion()}():\n"
                f'    raise NotImplementedError("To be implemented")'
            )
            return f"Requirement {req_info} is missing test case: \n\n{receipe}\n"

        if isinstance(excinfo.value, FailedTestCaseException):
            return (
                f"Requirement {req_info} test case failed. "
                "See corresponding test case for more details.\n"
            )

        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, self.req.line_no, f"RequirementTest: {self.name} {self.req.desc}"



import re
from typing import Dict, List

import lxml.etree  # type: ignore
import pytest

from ..exception import FailedTestCaseException
from ..schema import TestCase

JUNIT_PATTERN = r"(?P<req_id>[A-Z,0-9]+)( \[(?P<refs>[A-Z,0-9]+)\])?: (?P<desc>.*)"


class JUnitFile(pytest.File):
    def collect(self):
        with open(self.fspath) as f:
            xml_data = lxml.etree.parse(f)
            testsuites = xml_data.getroot()

            pattern = re.compile(
                self.session.config.getoption("nydok-junit-regex")
                or self.session.config.getini("nydok-junit-regex")
                or JUNIT_PATTERN
            )

            for testsuite in testsuites.findall("testsuite"):
                testcases = testsuite.findall("testcase")

                # We allow several results for a testcase within a single testsuite
                # If there are no failures, we select the first test case.
                # If there are failures, we select the first failure.
                chosen_testcases: Dict[str, JUnitItem] = {}
                for testcase in testcases:
                    name = testcase.attrib["name"]
                    classname = testcase.attrib.get("classname", "")
                    # file = testcase.attrib.get("file", "")
                    # line_no = testcase.attrib.get("line", "")

                    failures: List[str] = []
                    for failure in testcase.findall("failure"):
                        failures.append(failure.text)

                    skipped: List[str] = []
                    for skip in testcase.findall("skipped"):
                        skipped.append(skip.text)

                    # testcase_data = {
                    #     "path": self.fspath,
                    #     "name": name,
                    #     "file": file,
                    #     "line_no": line_no,
                    #     "failures": failures,
                    #     "skipped": skipped,
                    # }

                    # Parse out references
                    for re_result in re.finditer(pattern, name):
                        matches = re_result.groupdict()
                        refs = []
                        if matches.get("refs"):
                            refs = matches["refs"].split(",")
                        test_case = TestCase(
                            matches["req_id"].split(","),
                            None,
                            matches.get("desc"),
                            None,
                            classname,
                            None,
                            refs,
                            False,
                            not bool(failures),
                        )
                        if name not in chosen_testcases or (
                            failures and not chosen_testcases[name].failures
                        ):
                            chosen_testcases[name] = JUnitItem.from_parent(
                                self, name=name, test_case=test_case, failures=failures
                            )

                yield from chosen_testcases.values()


class JUnitItem(pytest.Item):
    def __init__(
        self,
        name,
        parent,
        test_case: TestCase,
        failures: List[str],
    ):
        super().__init__(name, parent)
        self.test_case: TestCase = test_case
        self.failures: List[str] = failures

    def runtest(self):
        if not self.test_case.passed:
            raise FailedTestCaseException()

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""

        if isinstance(excinfo.value, FailedTestCaseException):
            return f"JUnit testcase '{self.test_case.desc}' is reported as failed: \n" + "\n".join(
                self.failures
            )

        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, None, f"JUnitTestCase: {self.name}"



import os
from typing import Any, Dict, List, Optional
from urllib.parse import quote_plus

import requests  # type: ignore


def get_api_url() -> str:
    if not os.environ.get("GITLAB_URL"):
        raise RuntimeError("Environment variable GITLAB_URL is required, but not set")
    return os.environ["GITLAB_URL"] + "/api/v4"


def get_graphql_url() -> str:
    return os.environ["GITLAB_URL"] + "/api/graphql"


def get_gitlab_token() -> str:
    if not os.environ.get("GITLAB_TOKEN"):
        raise RuntimeError("Environment variable GITLAB_TOKEN is required, but not set")
    return os.environ["GITLAB_TOKEN"]


def _get_paginated_results(url: str, params: Dict[str, Any]) -> List[Any]:
    params = dict(params)
    params.setdefault("per_page", 100)

    def _get_results(next_page: Optional[str] = None):
        fetch_url = next_page if next_page else url
        resp = requests.get(
            fetch_url,
            params=params,
            headers={"Authorization": f"Bearer {get_gitlab_token()}"},
        )
        assert resp.status_code == 200, (
            f"Resource {fetch_url} returned status code {resp.status_code}.\n{resp.text}"
        )

        return resp.json(), resp.links.get("next", {}).get("url")

    api_results, next_page = _get_results()
    while next_page:
        next_jobs, next_page = _get_results(next_page)
        api_results.extend(next_jobs)
    return api_results


def get_commit_for_tag(repo_path: str, tag: str) -> str:
    url_encoded_repo_path = quote_plus(repo_path)

    fetch_url = get_api_url() + f"/projects/{url_encoded_repo_path}/repository/tags/{tag}"
    resp = requests.get(
        fetch_url,
        headers={"Authorization": f"Bearer {get_gitlab_token()}"},
    )
    assert resp.status_code == 200, (
        f"Resource {fetch_url} returned status code {resp.status_code}.\n{resp.text}"
    )
    return resp.json()["commit"]["id"]


def get_commits_for_ref(
    repo_path: str, to_ref: Optional[str] = None, from_ref: Optional[str] = None
) -> List[str]:
    """Fetches all commit ids for given branch"""

    url_encoded_repo_path = quote_plus(repo_path)

    params = {}
    if from_ref or to_ref:
        from_ref = from_ref or ""
        to_ref = to_ref or ""
        params["ref_name"] = f"{from_ref}...{to_ref}"

    api_commits = _get_paginated_results(
        get_api_url() + f"/projects/{url_encoded_repo_path}/repository/commits",
        {"ref_name": f"{from_ref}...{to_ref}" if from_ref else to_ref},
    )

    commits = [c["id"] for c in api_commits]
    return commits


def fetch_mergerequests(repo_path: str) -> List[Dict[str, str]]:
    url_encoded_repo_path = quote_plus(repo_path)

    api_mrs = _get_paginated_results(
        get_api_url() + f"/projects/{url_encoded_repo_path}/merge_requests",
        {
            "state": "merged",
        },
    )

    merge_requests = []
    for mr in api_mrs:
        merge_requests.append(
            {
                "merge_commit_sha": mr["merge_commit_sha"],
                "title": mr["title"],
                "merged_at": mr["merged_at"],
                "author": mr["author"]["name"],
                "approved_by": ",".join(sorted(r["name"] for r in mr["reviewers"])),
            }
        )
    return merge_requests


def get_pipeline_logs(
    repo_path: str, pipeline_id: int, job_names: Optional[List[str]] = None
) -> Dict[str, str]:
    url_encoded_repo_path = quote_plus(repo_path)

    api_jobs = _get_paginated_results(
        get_api_url() + f"/projects/{url_encoded_repo_path}/pipelines/{str(pipeline_id)}/jobs",
        {
            "scope": ["failed", "success"],
        },
    )

    logs: Dict[str, str] = {}
    for job in api_jobs:
        if job_names and job["name"] not in job_names:
            continue

        resp = requests.get(
            get_api_url() + f"/projects/{url_encoded_repo_path}/jobs/{str(job['id'])}/trace",
            headers={"Authorization": f"Bearer {get_gitlab_token()}"},
        )
        if resp.status_code != 200:
            raise RuntimeError(f"Failed to obtain log for job {job['id']}")

        logs[job["name"]] = resp.text

    return logs



class DuplicateRequirementException(Exception): ...


class FailedTestCaseException(Exception): ...


class MissingTestCaseException(Exception): ...


class RiskPriorityExceedsThresholdException(Exception): ...



import json
import sys
from pathlib import Path

import click

from .report import (
    create_codereview_report,
    create_pipeline_logs_report,
    create_risk_report,
    create_test_overview_report,
    create_test_case_report,
    create_traceability_matrix,
)
from .schema import Requirement, RiskAssessment, TestCase


@click.group()
def cli():
    pass


@cli.group()
def report():
    pass


def _load_json(path: Path):
    data = json.loads(path.read_text())
    return {
        "test_cases": {_id: TestCase(**test_case) for _id, test_case in data["test_cases"].items()},
        "requirements": {
            _id: Requirement(**requirement) for _id, requirement in data["requirements"].items()
        },
        "risk_assessments": {
            _id: RiskAssessment.from_dict(_id, risk_assessment)
            for _id, risk_assessment in data["risk_assessments"].items()
        },
    }


@report.command(help="Create traceability matrix.")
@click.argument("result", type=click.Path(exists=True))
@click.option("--base-prefix", type=str, help="Prefix to use as basis (filter).")
@click.option(
    "--categories",
    type=str,
    help="Categories for traceability (title,prefix). Repeat for multiple.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(writable=True),
    default="-",
    help="Output path (default: stdout).",
)
def traceability_matrix(result, base_prefix, categories, output):
    if categories:
        assert len(categories.split(",")) % 2 == 0, "Categories must be in pairs."
        categories = list(zip(categories.split(",")[::2], categories.split(",")[1::2]))

    data = _load_json(Path(result))
    report = create_traceability_matrix(
        data["requirements"].values(),
        data["test_cases"].values(),
        base_prefix=base_prefix,
        categories=categories,
    )
    if output == "-":
        sys.stdout.write(report + "\n")
    else:
        Path(output).write_text(report)


@report.command(help="Create risk assessment tables.")
@click.argument("result", type=click.Path(exists=True))
@click.option(
    "--output",
    "-o",
    type=click.Path(writable=True),
    default="-",
    help="Output path (default: stdout).",
)
def risk_assessment(result, output):
    data = _load_json(Path(result))
    report = create_risk_report(data["risk_assessments"])

    if output == "-":
        sys.stdout.write(report)
    else:
        Path(output).write_text(report)


@report.command(help="Create test overview table.")
@click.argument("result", type=click.Path(exists=True))
@click.option(
    "--output",
    "-o",
    type=click.Path(writable=True),
    default="-",
    help="Output path (default: stdout).",
)
def test_overview(result, output):
    data = _load_json(Path(result))
    report = create_test_overview_report(
        data["requirements"].values(),
        data["test_cases"].values(),
    )
    if output == "-":
        sys.stdout.write(report)
    else:
        Path(output).write_text(report)


@report.command(help="Create test case report.")
@click.argument("result", type=click.Path(exists=True))
@click.option(
    "--output",
    "-o",
    type=click.Path(writable=True),
    default="-",
    help="Output path (default: stdout).",
)
def test_cases(result, output):
    data = _load_json(Path(result))
    report = create_test_case_report(
        data["requirements"].values(),
        data["test_cases"].values(),
    )
    if output == "-":
        sys.stdout.write(report)
    else:
        Path(output).write_text(report)


@report.command(help="Create code review table.")
@click.option(
    "--repo-path",
    "-r",
    type=str,
    required=True,
    help="Gitlab repository path. E.g. my-group/my-project.",
)
@click.option(
    "--to-ref",
    "-c",
    type=str,
    default="main",
    help="Git ref for end of commit range. Default is default branch configured for repository.",
)
@click.option(
    "--from-ref",
    "-f",
    type=str,
    help="Git ref for start of commit range. Default is all history.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(writable=True),
    default="-",
    help="Output path (default: stdout).",
)
def code_review(repo_path, to_ref, from_ref, output):
    report = create_codereview_report(
        repo_path=repo_path,
        to_ref=to_ref,
        from_ref=from_ref,
    )
    if output == "-":
        sys.stdout.write(report)
    else:
        Path(output).write_text(report)


@report.command(help="Create pipeline logs report.")
@click.option(
    "--repo-path",
    "-r",
    type=str,
    required=True,
    help="Gitlab repository path. E.g. my-group/my-project.",
)
@click.option(
    "--pipeline-id",
    "-p",
    type=click.INT,
    required=True,
    help="Pipeline ID in Gitlab from which to retrieve job logs.",
)
@click.option(
    "--job-names",
    "-j",
    type=str,
    help="Job names in pipeline for only including certain jobs. Default is all jobs.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(writable=True),
    default="-",
    help="Output path (default: stdout).",
)
def pipeline_logs(repo_path, pipeline_id, job_names, output):
    job_names = job_names.split(",") if job_names else None

    report = create_pipeline_logs_report(
        repo_path=repo_path, pipeline_id=pipeline_id, job_names=job_names
    )
    if output == "-":
        sys.stdout.write(report)
    else:
        Path(output).write_text(report)



import dataclasses
import json
from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from typing import Any, Dict, List, Tuple


class DataclassJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        elif isinstance(o, Path):
            return str(o)
        elif isinstance(o, datetime):
            return o.isoformat()
        elif isinstance(o, date):
            return o.isoformat()
        return super().default(o)


# Ordered according to level of severity, don't change
RISK_ASSESSMENT_CATEGORIES = ("low", "medium", "high")

RISK_ASSESSMENT_CLASS = {
    # Given probability then severity
    "high": {"high": 1, "medium": 1, "low": 2},
    "medium": {"high": 1, "medium": 2, "low": 3},
    "low": {"high": 2, "medium": 3, "low": 3},
}

RISK_ASSESSMENT_PRIORITY = {
    # Given class then detectability
    1: {
        "high": "medium",
        "medium": "high",
        "low": "high",
    },
    2: {
        "high": "low",
        "medium": "medium",
        "low": "high",
    },
    3: {
        "high": "low",
        "medium": "low",
        "low": "medium",
    },
}


@dataclass
class TestCase:
    ids: List[str]
    testcase_id: str
    desc: str
    io: List[Tuple[Any, Any]]
    func_name: str
    func_src: str
    ref_ids: List[str]
    skip: bool
    passed: bool


@dataclass
class Requirement:
    id: str
    desc: str
    file_path: Path
    line_no: int
    ref_ids: List[str]


@dataclass
class RiskAssessment:
    id: str
    description: str
    consequence: str
    prior_probability: str
    prior_severity: str
    prior_detectability: str
    mitigation: str
    mitigation_requirement_ids: List[str]
    residual_probability: str
    residual_severity: str
    residual_detectability: str

    def __post_init__(self):
        risk_fields = [
            "prior_probability",
            "prior_severity",
            "prior_detectability",
            "residual_probability",
            "residual_severity",
            "residual_detectability",
        ]
        for field in risk_fields:
            if getattr(self, field) not in RISK_ASSESSMENT_CATEGORIES:
                raise ValueError(
                    f"Error while checking Risk Assessment {self.id}: {field} "
                    f"must be one of {RISK_ASSESSMENT_CATEGORIES}"
                )

        required_attribs = [
            "id",
            "description",
            "consequence",
            "mitigation",
        ]
        for attrib in required_attribs:
            if not getattr(self, attrib):
                raise ValueError(
                    f"Error while checking Risk Assessment {self.id}: {attrib} must be defined"
                )

    @property
    def prior_risk_priority(self) -> str:
        prior_risk_class = RISK_ASSESSMENT_CLASS[self.prior_probability][self.prior_severity]
        prior_risk_priority = RISK_ASSESSMENT_PRIORITY[prior_risk_class][self.prior_detectability]
        return prior_risk_priority

    @property
    def residual_risk_priority(self) -> str:
        residual_risk_class = RISK_ASSESSMENT_CLASS[self.residual_probability][
            self.residual_severity
        ]
        residual_risk_priority = RISK_ASSESSMENT_PRIORITY[residual_risk_class][
            self.residual_detectability
        ]
        return residual_risk_priority

    @classmethod
    def from_dict(cls, _id, risk_dict_object: Dict):
        """Populate a RiskAssessment from a dictionary object from a yaml-file."""
        _risk_dict_object = dict(risk_dict_object)
        _risk_dict_object.setdefault("id", _id)
        _risk_dict_object.setdefault("mitigation_requirement_ids", [])
        return cls(**_risk_dict_object)



import functools
import inspect
from typing import Any, Callable, List, Optional, Tuple, Union

from ..schema import TestCase

REQ_TESTCASE_TAG = "__req_testcase__"  # Attribute for storing metadata on the test functions


def testcase(
    req_id: Union[str, List[str]],
    testcase_id: Optional[str] = None,
    desc: str = "",
    io: Optional[List[Tuple[Any, Any]]] = None,
    ref_ids: Optional[List[str]] = None,
    skip=False,
) -> Callable:
    """Register test case for a corresponding requirement.

    Test data handling:

    When providing test data using the `io=` argument,
    the input and expected output can be documented in the test report.

    Example:

        @testcase("FR001", "TC001",
            io=[
                ((2, 4), 6),
                ((3, 3), 6)
            ]
        )
        def test_sum(io=None):
            for i, e in io:
                assert sum(i) == e

    Args:
        req_id: Requirement id.
        testcase_id: Override test case id. Default is to autogenerate one.
        io: Input/output test data in a tuple, or list of, (input, output).
        ref_ids: List of ids to other references. An id can be any string.
        skip: Whether checking test function result should be skipped. Defaults to False.

    """

    if io:
        if isinstance(io, tuple):
            io = [io]
        assert all(len(i) == 2 for i in io), (
            "io argument must be list of tuples [(input, expected_output)]"
        )

    if not isinstance(req_id, list):
        req_id = [req_id]

    def outer(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            # TODO: Check that func is a py.test function (test_)

            # Add io arguments to function
            if io:
                kwargs["io"] = io
            return func(*args, **kwargs)

        setattr(
            inner,
            REQ_TESTCASE_TAG,
            TestCase(
                req_id,
                testcase_id,
                desc,
                io,
                func.__name__,
                inspect.getsource(func),
                ref_ids,
                skip,
                False,
            ),
        )
        return inner

    return outer


testcase.__test__ = False  # type: ignore



import re
from typing import Dict, List, Optional, Set, Tuple

from .gitlab import fetch_mergerequests, get_commits_for_ref, get_pipeline_logs
from .schema import Requirement, RiskAssessment, TestCase


def _render_md_table(header, rows, max_col_width=100):
    def _md_table_row(cols: List[str], col_widths: List[int]):
        return " | ".join(
            # Pad each entry with col_widths
            [
                "{h:<{w}}".format(
                    h=cols[i],
                    w=col_widths[i],
                )
                for i in range(0, len(cols))
            ]
        )

    # Find max col size per column
    max_col_size: List[int] = []
    for idx in range(0, len(header)):
        max_col_size.append(min(max([len(r[idx]) for r in [header] + rows]), max_col_width))

    result = ""
    result += _md_table_row(header, max_col_size) + "\n"
    result += (
        _md_table_row(["-" * max_col_size[i] for i in range(0, len(header))], max_col_size) + "\n"
    )
    for row in rows:
        result += _md_table_row(row, max_col_size) + "\n"

    return result


def _get_ids_for_req(req: Requirement, req_impl: Optional[TestCase] = None) -> Set[str]:
    ref_ids: Set[str] = set()
    if req.ref_ids:
        ref_ids.update(req.ref_ids)
    if req_impl and req_impl.ref_ids:
        ref_ids.update(req_impl.ref_ids)

    return ref_ids


def _traverse_ids_for_req(
    req: Requirement,
    candidate_reqs: List[Requirement],
    candidate_impl: List[TestCase],
    prefix: str,
    seen: Optional[Set[str]] = None,
) -> Set[str]:
    result: Set[str] = set()

    if seen is None:
        seen = set()

    req_impl = next((i for i in candidate_impl if req.id in i.ids), None)
    ref_ids = _get_ids_for_req(req, req_impl)
    for _id in ref_ids:
        # Ensure no cycles, or we'll be stuck in an infinite loop
        if _id in seen:
            raise RuntimeError(f"Circular reference detected, {_id} has already been seen")
        seen.add(_id)

        if _id.startswith(prefix):
            result.add(_id)
        else:
            # Traverse referenced (child) requirements to search for more matches
            if referenced_req := next((r for r in candidate_reqs if r.id == _id), None):
                result.update(
                    _traverse_ids_for_req(
                        referenced_req, candidate_reqs, candidate_impl, prefix, seen
                    )
                )
    return result


def create_traceability_matrix(
    reqs: List[Requirement],
    test_cases: List[TestCase],
    categories: Optional[List[Tuple[str, str]]] = None,
    base_prefix: Optional[str] = None,
    max_col_width=100,
) -> str:
    """Creates a traceability matrix in markdown format.

    If categories is provided, the matrix will contain additional columns,
    one for each category, linking the different requirements together
    through their references.

    Args:
        reqs: Requirements to include.
        test_cases: TestCases for the Requirements
        categories: List of tuples, containing title and prefix that defines a category.
                    If none, all requirements are listed.
        base_prefix: Which requirement prefix to base the matrix upon. Required if
                     categories is provided.
        max_col_width: Maximum width of each column in the table.


    Returns:
        Traceability matrix table in Markdown format.
    """
    header = ["ID", "Description", "References", "Test case"]

    if categories:
        header = [c[0] for c in categories] + header

    rows: List[List[str]] = []
    for req in sorted(reqs, key=lambda x: x.id):
        # If base_prefix is given, table only relates to matching requirements
        if base_prefix and not req.id.startswith(base_prefix):
            continue

        req_test_case = next((i for i in test_cases if req.id in i.ids), None)

        category_entries: List[str] = []
        used_prefix_ids: Set[str] = set()
        if categories:
            for _, prefix in categories:
                prefix_ids = _traverse_ids_for_req(req, reqs, test_cases, prefix)
                if prefix_ids:
                    category_entries.append(", ".join(sorted(prefix_ids)))
                    used_prefix_ids.update(prefix_ids)
                else:
                    category_entries.append("-")

        rows.append(
            category_entries
            + [
                str(req.id),
                req.desc,
                # Remove the used prefixes from the list of references
                ", ".join(sorted(_get_ids_for_req(req, req_test_case) - used_prefix_ids)),
                str(
                    req_test_case.testcase_id if req_test_case and req_test_case.testcase_id else ""
                ),
            ]
        )

    report = _render_md_table(header, rows, max_col_width=max_col_width)

    return report


def create_test_overview_report(
    reqs: List[Requirement],
    test_cases: List[TestCase],
) -> str:
    report = ""

    # TestCases are repeated if duplicated IDs, use first one
    report_test_cases: Dict[str, TestCase] = {}
    for test_case in sorted(test_cases, key=lambda x: x.testcase_id):
        if test_case.testcase_id not in report_test_cases:
            report_test_cases[test_case.testcase_id] = test_case

    # Print summary table
    report += (
        _render_md_table(
            ["Test case", "Description", "Passed"],
            [
                [
                    tc.testcase_id,
                    tc.desc,
                    "Pass" if tc.passed else "**Fail**",
                ]
                for tc in report_test_cases.values()
            ],
        )
        + "\n\n"
    )

    return report


def create_test_case_report(
    reqs: List[Requirement],
    test_cases: List[TestCase],
) -> str:
    report = ""

    # TestCases are repeated if duplicated IDs, use first one
    report_test_cases: Dict[str, TestCase] = {}
    for test_case in sorted(test_cases, key=lambda x: x.testcase_id):
        if test_case.testcase_id not in report_test_cases:
            report_test_cases[test_case.testcase_id] = test_case

    # Print individual test cases
    for test_case in report_test_cases.values():
        impl_reqs = [s for s in reqs if s.id in test_case.ids]

        testdata_table = ""
        if test_case.io:
            rows: List[List[str]] = list()
            for io_entry in test_case.io:
                rows.append([str(io_entry[0]), str(io_entry[1])])

            testdata_table = _render_md_table(["Input", "Output"], rows)

        report += f"## {test_case.testcase_id}{': ' + test_case.desc if test_case.desc else ''}\n\n"

        report += _render_md_table(["ID", "Description"], [[req.id, req.desc] for req in impl_reqs])
        report += "\n\n"

        if testdata_table:
            report += f"### Test data:\n\n{testdata_table}\n"

        if test_case.func_src:
            report += "### Test case implementation:\n\n"
            report += f"```python\n{test_case.func_src}```\n\n"

    return report


RISK_ITEM_TEMPLATE = """
  <thead>
    <tr>
      <th class="nydok-risk-id" rowspan="2" style="text-align: center;">{_id}</th>
      <th class="nydok-risk-header nydok-risk-prior" colspan="3" style="text-align: center;">Prior</th>
      <th class="nydok-risk-header nydok-risk-residual" colspan="3" style="text-align: center;">Residual</th>
    </tr>
    <tr>
      <th class="nydok-risk-category nydok-risk-prior" style="text-align: center;">Prob.</th>
      <th class="nydok-risk-category nydok-risk-prior" style="text-align: center;">Severity</th>
      <th class="nydok-risk-category nydok-risk-prior" style="text-align: center;">Detect.</th>
      <th class="nydok-risk-category nydok-risk-residual" style="text-align: center;">Prob.</th>
      <th class="nydok-risk-category nydok-risk-residual" style="text-align: center;">Severity</th>
      <th class="nydok-risk-category nydok-risk-residual" style="text-align: center;">Detect.</th>
    </tr>
    <tr>
        {prior_probability}
        {prior_severity}
        {prior_detectability}
        {residual_probability}
        {residual_severity}
        {residual_detectability}
    </tr>
    <tr>
        <th>Risk priority</th>
        {prior_risk_priority}
        {residual_risk_priority}
    </tr>
  </thead>
  <tbody>
    <tr>
      <th class="nydok-risk-text-header" style="padding: 10px;">Description</th>
      <td class="nydok-risk-text" style="padding: 10px;" colspan="6">{description}</td>
    </tr>
    <tr>
      <th class="nydok-risk-text-header" style="padding: 10px;">Consequence</th>
      <td class="nydok-risk-text" style="padding: 10px;" colspan="6">{consequence}</td>
    </tr>
    <tr>
      <th class="nydok-risk-text-header" style="padding: 10px;">Mitigation</th>
      <td class="nydok-risk-text" style="padding: 10px;" colspan="6">{mitigation}</td>
    </tr>
  </tbody>
"""  # noqa: E501


def create_risk_report(
    risk_assessments: Dict[str, RiskAssessment],
) -> str:
    """
    Create a risk report from a list of risk assessments as a HTML table.

    The table is sorted by risk assessment ID.

    Args:
        risk_assessments: A dictionary of risk assessments, given their ID.

    Returns:
        A HTML table with the risk assessments.
    """
    report = ""

    def get_risk_item(item: str, desc: bool = False, _class: str = "", colspan: str = "1") -> str:
        """Creates a <td> for a risk item.

        Args:
            item: 'low', 'medium' or 'high'.
            desc: Whether color scale is descending, e.g. 'high' is green. Defaults to False.
            _class: HTML class to add to row. Defaults to "".
            colspan: HTML colspan attribute value. Defaults to "1".

        Returns:
            A HTML string with <td> element.
        """
        assert item in ["low", "medium", "high"]

        color = [
            # (font color, background color)
            ("#3CA03F", "#C6EFCE"),
            ("#9C6500", "#FFEB9C"),
            ("#FF2020", "#FFC7CE"),
        ]
        color_classes = ["green", "orange", "red"]

        # Used to lookup index of item in above color lists
        color_index = ["low", "medium", "high"]

        color_item_index = (
            list(reversed(color_index)).index(item) if desc else color_index.index(item)
        )
        font_color, background_color = color[color_item_index]
        item_class = color_classes[color_item_index]

        item_classes = ["nydok-risk-score", f"nydok-risk-score-{item_class}"]
        if _class:
            item_classes.append(_class)

        style = (
            f"color: {font_color}; background-color: {background_color}; "
            f"text-align: center; min-width: 65px;"
        )
        return (
            f'<td colspan={colspan} class="{" ".join(item_classes)}"'
            f' style="{style}">{item.capitalize()}</td>'
        )

    if risk_assessments:
        report += '<table class="nydok-risk-assessment">'
        for ra in sorted(risk_assessments.values(), key=lambda x: x.id):
            mitigation_text = ra.mitigation

            # Add requirement ids to mitigation text if present
            if ra.mitigation_requirement_ids:
                if not mitigation_text.endswith("."):
                    mitigation_text += "."
                mitigation_text += (
                    " Mitigation requirement IDs: "
                    + ", ".join(sorted(ra.mitigation_requirement_ids))
                    + "."
                )

            report += RISK_ITEM_TEMPLATE.format(
                _id=ra.id,
                description=ra.description,
                consequence=ra.consequence,
                mitigation=mitigation_text,
                prior_probability=get_risk_item(ra.prior_probability, _class="nydok-risk-prior"),
                prior_severity=get_risk_item(ra.prior_severity, _class="nydok-risk-prior"),
                prior_detectability=get_risk_item(
                    ra.prior_detectability, desc=True, _class="nydok-risk-prior"
                ),
                prior_risk_priority=get_risk_item(
                    ra.prior_risk_priority, colspan="3", _class="nydok-risk-prior"
                ),
                residual_probability=get_risk_item(
                    ra.residual_probability, _class="nydok-risk-residual"
                ),
                residual_severity=get_risk_item(ra.residual_severity, _class="nydok-risk-residual"),
                residual_detectability=get_risk_item(
                    ra.residual_detectability, desc=True, _class="nydok-risk-residual"
                ),
                residual_risk_priority=get_risk_item(
                    ra.residual_risk_priority, colspan="3", _class="nydok-risk-residual"
                ),
            )
        report += "</table>"

    return report


def create_codereview_report(
    repo_path: str,
    to_ref: Optional[str] = None,
    from_ref: Optional[str] = None,
) -> str:
    """Create a report of code changes and their approval status since a given commit.

    Args:
        repo_path: Path to the repository to generate the report for.
        to_ref: The commit to generate the report for. If not given, repository
            HEAD is used.
        from_ref: The commit to generate the report from. If not given, all
            commits up until `to_ref` are considered.

    Returns:
        A markdown formatted string containing the report.
    """

    # Get commits from to_ref so we can figure out which merge requests
    # were merged into the commit range. If a from_ref is given, only merge requests
    # merged since the from_ref are considered.

    commits = get_commits_for_ref(repo_path, to_ref, from_ref=from_ref)

    since_prev_text = f" since version {from_ref}" if from_ref else ""
    report = f"## Approved code changes{since_prev_text}\n\n"

    mrs = fetch_mergerequests(repo_path)

    # Filter the merge requests against the commits
    to_include_mrs = [mr for mr in mrs if mr["merge_commit_sha"] in set(commits)]

    report += _render_md_table(
        ["Commit", "Title", "Date", "Author", "Approver(s)"],
        [
            [
                mr["merge_commit_sha"][:7],
                mr["title"],
                mr["merged_at"][:10],
                mr["author"],
                mr["approved_by"],
            ]
            for mr in to_include_mrs
        ],
    )

    return report


def create_pipeline_logs_report(
    repo_path: str, pipeline_id: int, job_names: Optional[List[str]] = None
) -> str:
    logs = get_pipeline_logs(repo_path, pipeline_id, job_names=job_names)

    def ansi_to_plain(ansi):
        return re.sub(r"\x1B\[[0-9;]*[ABCDEFGHJKSTfmnsulh]", "", ansi)

    report = ""
    for job_name, job_log in sorted(logs.items()):
        job_log = ansi_to_plain(job_log.replace("\n", "<br>"))
        report += f"## {job_name}\n\n"
        report += f'<div class="codehilite"><pre><code>\n{job_log}\n</code></pre></div>\n\n'

    return report



import json
from pathlib import Path
from typing import Dict

import yaml

from ..exception import (
    DuplicateRequirementException,
    RiskPriorityExceedsThresholdException,
)
from ..schema import (
    DataclassJsonEncoder,
    Requirement,
    RiskAssessment,
    TestCase,
    RISK_ASSESSMENT_CATEGORIES,
)


class SpecsManager:
    def __init__(self):
        self.test_cases: Dict[str, TestCase] = {}
        self.requirements: Dict[str, Requirement] = {}
        self.testcase_prefix = "TC"
        self.testcase_no = 1
        self.risk_assessments: Dict[str, RiskAssessment] = {}

    def enable_risk_assessment(self, path: str):
        with open(path, "r") as file:
            for _id, risk in yaml.full_load(file).items():
                self.risk_assessments[_id] = RiskAssessment.from_dict(_id, risk)

    def _format_testcase_id(self) -> str:
        return f"{self.testcase_prefix}{self.testcase_no:03}"

    def add_requirement(self, req: Requirement) -> None:
        self.requirements[req.id] = req

    def _add_test_case(self, req_id: str, test_case: TestCase) -> None:
        if req_id in self.test_cases:
            existing = self.test_cases[req_id]
            raise DuplicateRequirementException(
                f"TestCase for id {req_id} is already added by {existing.func_name}."
            )
        self.test_cases[req_id] = test_case

    def add_test_case(self, test_case: TestCase) -> None:
        if not test_case.skip and not test_case.testcase_id:
            test_case.testcase_id = self._format_testcase_id()
            self.testcase_no += 1

        for _id in test_case.ids:
            self._add_test_case(_id, test_case)

    def get_test_case(self, req_id: str) -> TestCase:
        return self.test_cases[req_id]

    def has_test_case(self, req_id: str) -> bool:
        return req_id in self.test_cases

    def requirement_passed(self, req_id: str) -> bool:
        return self.test_cases[req_id].passed

    def check_risk_assessment(self, permissable_risk_priority: str):
        for ra_id, ra in self.risk_assessments.items():
            if RISK_ASSESSMENT_CATEGORIES.index(
                ra.residual_risk_priority
            ) > RISK_ASSESSMENT_CATEGORIES.index(permissable_risk_priority):
                raise RiskPriorityExceedsThresholdException(
                    f"Risk assessment {ra_id}: Residual risk "
                    f"priority '{ra.residual_risk_priority}' "
                    f"exceeds permissable risk priority '{permissable_risk_priority}'."
                )

            # Check requirement ids
            for req_id in ra.mitigation_requirement_ids:
                if req_id not in self.requirements:
                    raise ValueError(
                        f"Risk assessment {ra_id}: mitigation_requirement_id '{req_id}' not found."
                    )

    def to_json(self, path: Path):
        data = json.dumps(
            {
                "test_cases": self.test_cases,
                "requirements": self.requirements,
                "risk_assessments": self.risk_assessments,
            },
            cls=DataclassJsonEncoder,
            indent=4,
        )
        path.write_text(data)


specs_manager = SpecsManager()



import re
from pathlib import Path
from typing import Dict

import pytest

from .junit import JUNIT_PATTERN, JUnitFile, JUnitItem
from .specification import DEFAULT_REQ_PATTERN, ReqItem, SpecFile
from .specsmanager import specs_manager
from .testcase import REQ_TESTCASE_TAG

RE_NEW_LINE = re.compile(r"\n")


def _get_option_or_ini(config, setting, default=None):
    return config.getoption(setting) or config.getini(setting) or default


# Pytest lifecycle hooks, in order of occurrence:
def pytest_addoption(parser):
    """Pytest specific hook that is triggered upon setting up the command argument parser."""
    group = parser.getgroup("nydok", "Specification testing")

    def add_opt_ini(opt_name: str, help: str):
        group.addoption(f"--{opt_name}", dest=opt_name, help=help, type=str)
        parser.addini(opt_name, help, type="string")

    add_opt_ini(
        "nydok-specs-regex", f"Regex for specification parsing. Default: '{DEFAULT_REQ_PATTERN}'"
    )
    add_opt_ini("nydok-junit-regex", f"Regex for JUnit name parsing. Default: '{JUNIT_PATTERN}'")
    add_opt_ini("nydok-risk-assessment", "Input path for risk assessment.")
    add_opt_ini("nydok-risk-priority-threshold", "Threshold for risk priority. Default: 'low'")
    add_opt_ini("nydok-output", "Path for nydok json results file.")


def pytest_configure(config):
    """Pytest specific hook that is triggered after options are parsed."""
    if ra_file := _get_option_or_ini(config, "nydok-risk-assessment"):
        specs_manager.enable_risk_assessment(ra_file)


def pytest_collect_file(parent, path):
    """Pytest specific hook that is triggered for evaluating inclusion of a file in the tests."""
    if str(path).endswith(".spec.md"):
        return SpecFile.from_parent(parent, fspath=path)

    if str(path).endswith(".junit.xml"):
        return JUnitFile.from_parent(parent, fspath=path)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    """Pytest specific hook that is triggered for each test being run."""

    def is_pytest_with_testcase(item):
        return isinstance(item, pytest.Function) and hasattr(item.function, REQ_TESTCASE_TAG)

    test_case = None

    if isinstance(item, JUnitItem):
        test_case = item.test_case
    elif is_pytest_with_testcase(item):
        test_case = getattr(item.function, REQ_TESTCASE_TAG)

    if test_case:
        specs_manager.add_test_case(test_case)

    # Run the test
    result = yield

    # Update TestCase.passed if the test passed
    if is_pytest_with_testcase(item) and not result.excinfo:
        test_case.passed = True


def pytest_collection_finish(session):
    """Runs after py.test is done collecting test Items.

    We need to ensure that all ReqItems runs last,
    in order to make sure that corresponding TestCaseItems can be found.
    """

    def sort_req_last(item):
        # Sort function for sorting ReqItems last,
        # with the Python functions next.
        # Any other items are run first.
        if isinstance(item, ReqItem):
            return 2
        if isinstance(item, pytest.Function):
            return 1
        return 0

    session.items = sorted(session.items, key=sort_req_last)

    req_items = [i for i in session.items if isinstance(i, ReqItem)]
    seen_names: Dict[str, ReqItem] = dict()
    for req_item in req_items:
        if req_item.name in seen_names:
            req_item.duplicate_item = seen_names[req_item.name]
        seen_names[req_item.name] = req_item

    # Ensure that all risk assessments' risk priority are within the specified threshold
    permissable_risk_priority = _get_option_or_ini(
        session.config, "nydok-risk-priority-threshold", default="low"
    )
    specs_manager.check_risk_assessment(permissable_risk_priority)


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session):
    """Pytest specific hook that is triggered as the pytest session is complete."""

    outcome = yield
    outcome.get_result()

    # Optionally write the results to a file
    if output := _get_option_or_ini(session.config, "nydok-output"):
        specs_manager.to_json(Path(output))
